;; -*-Mode: Emacs-Lisp;-*-
;; PRCS - The Project Revision Control System
;; Copyright (C) 1997  Josh MacDonald
;;
;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License
;; as published by the Free Software Foundation; either version 2
;; of the License, or (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software
;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;;
;; $Id: prcs.el 1.3 Fri, 16 May 1997 17:51:28 -0700 jmacd $

;; PRCS project file parser

(require 'lisp-mode)
(require 'emerge)
(require 'cl)
(require 'add-log)

(defvar prcs-auto-add-changelog t
  "If non nil and the project contains a file matching named
ChangeLog in its root directory, prcs-mode will automatically insert
the contents of New-Version-Log into the ChangeLog.")

(defvar prcs-ask-about-save t
  "If not nil, PRCS checkin asks which buffers to save before checking in.
Otherwise, it saves all modified buffers without asking.")

(defvar prcs-program-name "prcs"
  "Name of the prcs executable.")

(defvar prcs-output-buffer-name "*prcs-output*"
  "Name of the prcs output buffer.")

(defvar prcs-extra-checkin-args nil
  "Extra arguments passed to checkin.")

(defvar prcs-extra-diff-args '("-c")
  "Extra arguments passed to checkin.")

(defconst prcs::prj-ignore-regex "\\(\\(;.*\\)\\|[ \t\n]\\)+"
  "Whitespace and comments")
(defconst prcs::prj-string-regex "\"\\([^\\\"]\\|\\(\\\\\\(.\\|\n\\)\\)\\)*\""
  "String literals")
(defconst prcs::prj-label-regex  "\\([^\\\"() \t\n]\\|\\(\\\\\\(.\\|\n\\)\\)\\)+"
  "Labels")

(or (assq 'prcs-mode font-lock-defaults-alist)
    (setq font-lock-defaults-alist
	  (cons
	   (cons 'prcs-mode
		 '(prcs-font-lock-keywords
		   nil
		   t
		   (("+-*/.<>=!?$%_&~^:" . "w"))
		   beginning-of-defun
		   (font-lock-comment-start-regexp . ";")
		   (font-lock-mark-block-function . mark-defun))
		 )
	   font-lock-defaults-alist
	   )
	  )
    )

(defconst prcs-font-lock-keywords
  (list
   '("(\\(Files\\|Created-By-Prcs-Version\\|Project-Description\\|Project-Version\\|Parent-Version\\|Version-Log\\|New-Version-Log\\|Checkin-Time\\|Checkin-Login\\|Populate-Ignore\\|Project-Keywords\\|Merge-Parents\\|New-Merge-Parents\\)" 1 font-lock-function-name-face)
   '("\\<:\\sw+\\>" 0 font-lock-reference-face prepend)
   )
 "PRCS keywords...")

(defun prcs::parse-prj-file (buffer)
  (message "Parsing project file...")
  (set-buffer buffer)  (goto-char (point-min))
  (let ((prj-sexps nil))
    (while (< (point) (point-max))
      (let ((s  (prcs::read-sexp)))
	(if s (setq prj-sexps (cons s prj-sexps)))))
    (nreverse prj-sexps)))

(defun prcs::read-sexp()
  (if (looking-at prcs::prj-ignore-regex)
      (goto-char (match-end 0)))
  (let ((beg-point (point))
	(cur-char (following-char)))
    (cond ((eq cur-char 0) nil)  ; EOF
	  ((eq cur-char ?\()     ; read a list
	   (forward-char)
	   (let ((l (prcs::read-list)))
	     (list l beg-point (point))))
	  ((eq cur-char ?\))     ; unbalanced parens
	   (error "unbalanced parens"))
	  ((eq cur-char ?\")     ; start a string
	   (if (not (looking-at prcs::prj-string-regex))
	       (error "unterminated string literal"))
	   (goto-char (match-end 0))
	   (list (buffer-substring-no-properties beg-point (point)) beg-point (point)))
	  (t                     ; start a label
	   (if (not (looking-at prcs::prj-label-regex))
	       (error "illegal label"))
	   (goto-char (match-end 0))
	   (list (buffer-substring-no-properties beg-point (point)) beg-point (point))))))

(defun prcs::read-list()
  (if (looking-at prcs::prj-ignore-regex)
      (goto-char (match-end 0)))
  (let ((cur-char (following-char)))
    (cond ((eq cur-char ?\))      ; end a list
	   (forward-char)
	   nil)
	  ((eq cur-char 0)        ; EOF in list
	   (error "EOF while scanning prj file"))
	  (t                      ; another elt
	   (let ((s (prcs::read-sexp)))
	     (cons s (prcs::read-list)))))))

;; PRCS-Emerge
;;
;; These functions allow the script 'emerge', distributed with PRCS, to
;; invoke emerge during merge.  It uses a recursive edit.  Probably should
;; know what you're doing.

(defun prcs::emerge-files (file-A file-B file-out quit-hooks)
  "Run Emerge on two files."
  (emerge-files-internal file-A file-B nil quit-hooks file-out))

(defun prcs::emerge-files-with-ancestor (file-A file-B file-ancestor file-out quit-hooks)
  "Run Emerge on two files, giving another file as the ancestor."
  (emerge-files-with-ancestor-internal file-A file-B file-ancestor nil quit-hooks file-out))

(defun prcs::emerge(work com sel out)
  (message "prcs::emerge with files %s %s %s %s" work com sel out)
  (let (quit-hooks)
    (add-hook 'quit-hooks (function (lambda () (exit-recursive-edit))))
    (add-hook 'quit-hooks (` (lambda () (emerge-files-exit (, out)))))
    (if (equal com "/dev/null")
	(prcs::emerge-files work sel out quit-hooks)
      (prcs::emerge-files-with-ancestor work sel com out quit-hooks))
    (recursive-edit)
    (message "prcs::emerge finished")
    )
  )

;; This is based on lisp-mode.el

;;; Code:

(defvar prcs-mode-syntax-table lisp-mode-syntax-table "")
(defvar prcs-mode-abbrev-table nil "")

(define-abbrev-table 'prcs-mode-abbrev-table ())

(defun prcs-mode-variables ()
  (setq local-abbrev-table prcs-mode-abbrev-table)
  (make-local-variable 'paragraph-start)
  (setq paragraph-start (concat page-delimiter "\\|$" ))
  (make-local-variable 'paragraph-separate)
  (setq paragraph-separate paragraph-start)
  (make-local-variable 'paragraph-ignore-fill-prefix)
  (setq paragraph-ignore-fill-prefix t)
  (make-local-variable 'fill-paragraph-function)
  (setq fill-paragraph-function 'lisp-fill-paragraph)
  ;; Adaptive fill mode gets in the way of auto-fill,
  ;; and should make no difference for explicit fill
  ;; because lisp-fill-paragraph should do the job.
  (make-local-variable 'adaptive-fill-mode)
  (setq adaptive-fill-mode nil)
  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'lisp-indent-line)
  (make-local-variable 'indent-region-function)
  (setq indent-region-function 'lisp-indent-region)
  (make-local-variable 'parse-sexp-ignore-comments)
  (setq parse-sexp-ignore-comments t)
  (make-local-variable 'outline-regexp)
  (setq outline-regexp ";;; \\|(....")
  (make-local-variable 'comment-start)
  (setq comment-start ";")
  (make-local-variable 'comment-start-skip)
  ;; Look within the line for a ; following an even number of backslashes
  ;; after either a non-backslash or the line beginning.
  (make-local-variable 'max-lisp-eval-depth)
  (setq max-lisp-eval-depth 10000)
  (make-local-variable 'max-specpdl-size)
  (setq max-specpdl-size 10000)
  (setq comment-start-skip "\\(\\(^\\|[^\\\\\n]\\)\\(\\\\\\\\\\)*\\);+ *")
  (make-local-variable 'comment-column)
  (setq comment-column 40)
  (make-local-variable 'comment-indent-function)
  (setq comment-indent-function 'lisp-comment-indent))

(defvar prcs-mode-map nil "")

(if prcs-mode-map
    ()
  (let ((map (make-sparse-keymap)))
    (define-key map "\C-c\C-c" 'prcs::prcs-checkin)
    (define-key map "\C-c\C-d" 'prcs::prcs-diff)
    (setq prcs-mode-map (nconc map shared-lisp-mode-map))
    )
  "Keymap for `prcs-mode'.")

(defvar prcs-mode-hook '() "Hooks for prcs-mode.")

(defun prcs-mode ()
  "Major mode for editing PRCS project files.
Commands:
Delete converts tabs to spaces as it moves back.
Blank lines separate paragraphs.  Semicolons start comments.
\\{prcs-mode-map}
Entry to this mode calls the value of `prcs-mode-hook'
if that value is non-nil.
C-c C-c runs PRCS checkin."
  (interactive)
  (kill-all-local-variables)
  (use-local-map prcs-mode-map)
  (set-syntax-table prcs-mode-syntax-table)
  (setq major-mode 'prcs-mode)
  (setq mode-name "PRCS")
  (prcs-mode-variables)
  (run-hooks 'prcs-mode-hook))

(put 'Files 'lisp-indent-function 0)
(put 'Created-By-Prcs-Version 'lisp-indent-function 0)
(put 'Project-Description 'lisp-indent-function 0)
(put 'Project-Version 'lisp-indent-function 0)
(put 'Parent-Version 'lisp-indent-function 0)
(put 'Version-Log 'lisp-indent-function 0)
(put 'New-Version-Log 'lisp-indent-function 0)
(put 'Checkin-Time 'lisp-indent-function 0)
(put 'Checkin-Login 'lisp-indent-function 0)
(put 'Populate-Ignore 'lisp-indent-function 0)
(put 'Project-Keywords 'lisp-indent-function 0)
(put 'Merge-Parents 'lisp-indent-function 0)
(put 'New-Merge-Parents 'lisp-indent-function 0)

(defmacro prcs::expr-val (x)
  (list 'car x))
(defmacro prcs::expr-start (x)
  (list 'car (list 'cdr x)))
(defmacro prcs::expr-end (x)
  (list 'car (list 'cdr (list 'cdr x))))

(defun prcs::prcs-checkin (&optional noact)
  (interactive "P")
  (let* ((prj (save-excursion (prcs::parse-prj-file (current-buffer))))
	 (log (prcs::find-expr prj "New-Version-Log")))
    (when (not log)
      (goto-char (point-min))
      (error "Can't find New-Version-Log"))
    (when (not (= (length log) 2))
      (goto-char (prcs::expr-start (car log)))
      (error "Invalid New-Version-Log"))
    (let* ((logstart (prcs::expr-start (cadr log)))
	   (logend (prcs::expr-end (cadr log))))
      (when (= logend (+ logstart 2))
	(goto-char (+ logstart 1))
	(if (y-or-n-p "No version log entered, enter one first? ")
	    (error "Checkin aborted."))
	)
      (if (and prcs-auto-add-changelog (not noact))
	  (prcs::add-changelog prj (buffer-substring-no-properties (+ logstart 1)
								   (- logend 1))))
      (save-some-buffers (not prcs-ask-about-save) nil)
      (cd-absolute (or (file-name-directory (buffer-file-name)) "."))
      (let ((exited (apply 'call-process
			   (append (list prcs-program-name
					 nil
					 (get-buffer-create prcs-output-buffer-name)
					 nil
					 "checkin"
					 (if noact
					     "-fln"
					   "-fl"))
				   prcs-extra-checkin-args
				   (list (file-name-nondirectory (buffer-file-name)))))))
	(let ((name (buffer-file-name)))
	  (kill-buffer (current-buffer))
	  (find-file name)
	  )
	(switch-to-buffer-other-window (get-buffer-create prcs-output-buffer-name))
	(if (not (= exited 0))
	    (message "PRCS exited with non-zero status--failed"))
	)
      )
    )
  )

(defun prcs::find-expr (prj name)
  (cond ((null prj) ())
	((and (consp (prcs::expr-val (car prj)))
	      (equal name (prcs::expr-val (car (prcs::expr-val (car prj))))))
	 (prcs::expr-val (car prj)))
	(t (prcs::find-expr (cdr prj) name))))

(defun prcs::add-changelog (prj message)
  (let ((files (prcs::find-expr prj "Files")))
    (when (not files)
      (goto-char (point-min))
      (error "Can't find Files"))
    (let ((changelog (prcs::find-expr (cdr files) (file-name-nondirectory (change-log-name)))))
      (when changelog
	(save-window-excursion
	  (message "Adding a ChangeLog entry...")
	  (add-change-log-entry)
	  (insert message)
	  (save-buffer)
	  )
	)
      )
    )
  )

(defun prcs::prcs-diff (&optional arg)
  (interactive)
  (save-some-buffers (not prcs-ask-about-save) nil)
  (cd-absolute (or (file-name-directory (buffer-file-name)) "."))
  (let ((exited (apply 'call-process
		       (append (list prcs-program-name
				     nil
				     (get-buffer-create prcs-output-buffer-name)
				     nil
				     "diff"
				     "-f"
				     (file-name-nondirectory (buffer-file-name)))
			       (if prcs-extra-diff-args
				   (append (list "--") prcs-extra-diff-args)
				 nil)))))
    (switch-to-buffer-other-window (get-buffer-create prcs-output-buffer-name))
    (cond ((= exited 2)
	   (message "PRCS exited with error status--failed"))
	  ((= exited 1)
	   (message "There were differences."))
	  (t
	   (message "No differences.")))
    )
  )

(provide 'prcs)
